@@ -242,6 +242,13 @@ insert_route(struct babel_route *route)
     return route;
 }
 
+static void
+destroy_route(struct babel_route *route)
+{
+    free(route->channels);
+    free(route);
+}
+
 void
 flush_route(struct babel_route *route)
 {
@@ -267,7 +274,7 @@ flush_route(struct babel_route *route)
     if(route == routes[i]) {
         routes[i] = route->next;
         route->next = NULL;
-        free(route);
+        destroy_route(route);
 
         if(routes[i] == NULL) {
             if(i < route_slots - 1)
@@ -275,6 +282,7 @@ flush_route(struct babel_route *route)
                         (route_slots - i - 1) * sizeof(struct babel_route*));
             routes[route_slots - 1] = NULL;
             route_slots--;
+            VALGRIND_MAKE_MEM_UNDEFINED(routes + route_slots, sizeof(struct route *));
         }
 
         if(route_slots == 0)
@@ -287,7 +295,7 @@ flush_route(struct babel_route *route)
             r = r->next;
         r->next = route->next;
         route->next = NULL;
-        free(route);
+        destroy_route(route);
     }
 
     if(lost)
@@ -376,7 +384,7 @@ route_stream(int which)
     if(!check_specific_first())
         fprintf(stderr, "Invariant failed: specific routes first in RIB.\n");
 
-    stream = malloc(sizeof(struct route_stream));
+    stream = calloc(1, sizeof(struct route_stream));
     if(stream == NULL)
         return NULL;
 
@@ -619,10 +627,9 @@ route_interferes(struct babel_route *route, struct interface *ifp)
             return 1;
         if(diversity_kind == DIVERSITY_CHANNEL) {
             int i;
-            for(i = 0; i < DIVERSITY_HOPS; i++) {
-                if(route->channels[i] == 0)
-                    break;
-                if(channels_interfere(ifp->channel, route->channels[i]))
+            for(i = 0; i < route->channels_len; i++) {
+                if(route->channels[i] != 0 &&
+                   channels_interfere(ifp->channel, route->channels[i]))
                     return 1;
             }
         }
@@ -909,10 +916,25 @@ update_route(const unsigned char *id,
             route->time = now.tv_sec;
         route->seqno = seqno;
 
-        memset(&route->channels, 0, sizeof(route->channels));
-        if(channels_len > 0)
-            memcpy(&route->channels, channels,
-                   MIN(channels_len, DIVERSITY_HOPS));
+        if(channels_len == 0) {
+            free(route->channels);
+            route->channels = NULL;
+            route->channels_len = 0;
+        } else {
+            if(channels_len != route->channels_len) {
+                unsigned char *new_channels =
+                    realloc(route->channels, channels_len);
+                if(new_channels == NULL) {
+                    perror("malloc(channels)");
+                    /* Truncate the data. */
+                    channels_len = MIN(channels_len, route->channels_len);
+                } else {
+                    route->channels = new_channels;
+                }
+            }
+            memcpy(route->channels, channels, channels_len);
+            route->channels_len = channels_len;
+        }
 
         change_route_metric(route,
                             refmetric, neighbour_cost(neigh), add_metric);
@@ -938,7 +960,7 @@ update_route(const unsigned char *id,
                 return NULL;
         }
 
-        route = malloc(sizeof(struct babel_route));
+        route = calloc(1, sizeof(struct babel_route));
         if(route == NULL) {
             perror("malloc(route)");
             return NULL;
@@ -955,16 +977,19 @@ update_route(const unsigned char *id,
         route->hold_time = hold_time;
         route->smoothed_metric = MAX(route_metric(route), INFINITY / 2);
         route->smoothed_metric_time = now.tv_sec;
-        route->installed = 0;
-        memset(&route->channels, 0, sizeof(route->channels));
-        if(channels_len > 0)
-            memcpy(&route->channels, channels,
-                   MIN(channels_len, DIVERSITY_HOPS));
+        if(channels_len > 0) {
+            route->channels = malloc(channels_len);
+            if(route->channels == NULL) {
+                perror("malloc(channels)");
+            } else {
+                memcpy(route->channels, channels, channels_len);
+            }
+        }
         route->next = NULL;
         new_route = insert_route(route);
         if(new_route == NULL) {
             fprintf(stderr, "Couldn't insert route.\n");
-            free(route);
+            destroy_route(route);
             return NULL;
         }
         local_notify_route(route, LOCAL_ADD);
@@ -1068,7 +1093,6 @@ retract_neighbour_routes(struct neighbour *neigh)
             }
             r = r->next;
         }
-        i++;
     }
 }
 
