@@ -104,9 +104,7 @@ add_xroute(unsigned char prefix[16], unsigned char plen,
     if(numxroutes >= maxxroutes) {
         struct xroute *new_xroutes;
         int n = maxxroutes < 1 ? 8 : 2 * maxxroutes;
-        new_xroutes = xroutes == NULL ?
-            malloc(n * sizeof(struct xroute)) :
-            realloc(xroutes, n * sizeof(struct xroute));
+        new_xroutes = realloc(xroutes, n * sizeof(struct xroute));
         if(new_xroutes == NULL)
             return -1;
         maxxroutes = n;
@@ -140,11 +138,10 @@ struct
 xroute_stream *
 xroute_stream()
 {
-    struct xroute_stream *stream = malloc(sizeof(struct xroute_stream));
+    struct xroute_stream *stream = calloc(1, sizeof(struct xroute_stream));
     if(stream == NULL)
         return NULL;
 
-    stream->index = 0;
     return stream;
 }
 
